/*
    mandelbmp - render the Mandelbrot set or Julia sets as BMP files
    Copyright (C) 2022 cas <cas@sdf.org>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "bmp.h"

#include <inttypes.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>
#include <stdint.h>

/*
 * Code for bitmap files. Modeled after german wikipedia article on
 * the BMP format, very incomplete, only does what I need here.
 * 
 * https://de.wikipedia.org/wiki/Windows_Bitmap
 * 
 * */

int write_BMP(char* fname, BMP* bmp){
  int fd = open(fname, O_CREAT | O_TRUNC | O_WRONLY, 0666);

  if(fd == -1){
    return -1;
  }
  
  write(fd, bmp -> file_header, 14);
  write(fd, bmp -> info_header, 40);
  
  // we're not writing any color tables
  
  lseek(fd, bmp -> file_header -> bfOffBits, SEEK_SET); // go to image data
  
  write(fd, bmp -> image_data, (bmp -> info_header) -> biSizeImage);
  
  close(fd);
}

void set_rgb(unsigned int x, unsigned int y, uint8_t r, uint8_t g, uint8_t b, BMP* bmp){
  unsigned int line_bytes = (bmp -> info_header -> biWidth) * 3;
  if(line_bytes % 4 != 0)
    line_bytes += - (line_bytes % 4) + 4;
  
  if(x < bmp -> info_header -> biWidth &&  y < bmp -> info_header -> biHeight){
    uint8_t* b_data = &((bmp -> image_data)[line_bytes * y + 3 * x]);
    uint8_t* g_data = &((bmp -> image_data)[line_bytes * y + 3 * x + 1]);
    uint8_t* r_data = &((bmp -> image_data)[line_bytes * y + 3 * x + 2]);
    
    *b_data = b;
    *g_data = g;
    *r_data = r;
  }
}

BMP* create_BMP(unsigned int w, unsigned int h){
   BMP* bmp = malloc(sizeof(BMP));
  if(bmp == NULL)
    return NULL;
  
  BMPFileHeader* file_header = malloc(sizeof(BMPFileHeader));
  BMPInfoHeader* info_header = malloc(sizeof(BMPInfoHeader));
  
  if(info_header == NULL){
    free(file_header);
    goto fail;
  }
  
  if(info_header == NULL)
    goto fail;
  
  file_header -> bfType = 0x4D42; // BM
  
  info_header -> biWidth  = w;
  info_header -> biHeight = h;
  // calculate line length and image size
  unsigned int line_bytes = w * 3;
  if(line_bytes % 4 != 0)
    line_bytes += - (line_bytes % 4) + 4;
  
  info_header -> biSizeImage     = line_bytes * h;
  info_header -> biClrUsed       = 0;
  info_header -> biClrImportant  = 0;
  info_header -> biBitCount      = 24;
  info_header -> biPlanes        = 1;
  info_header -> biCompression   = BI_RGB;
  info_header -> biXPelsPerMeter = 2835;
  info_header -> biYPelsPerMeter = 2835;
  info_header -> biSize          = sizeof(BMPInfoHeader);
  
  file_header -> bfOffBits = sizeof(BMPInfoHeader) + sizeof(BMPFileHeader);
  file_header -> bfSize    = file_header -> bfOffBits + line_bytes * h;
  
  bmp -> file_header = file_header;
  bmp -> info_header = info_header;
  
  bmp -> image_data = malloc(sizeof(uint32_t) * info_header -> biSizeImage);
 
  return bmp;
  
fail:
  free(bmp);
  free(file_header);
  return NULL;
}

void free_BMP(BMP* bmp){
  free(bmp -> info_header);
  free(bmp -> file_header);
  free(bmp -> image_data);
  free(bmp);
}
